import { mapValues } from './utils';
export function getNodes(node) {
    const nodes = Object.keys(node.states).reduce((accNodes, stateKey) => {
        const subState = node.states[stateKey];
        const subNodes = getNodes(node.states[stateKey]);
        accNodes.push(subState, ...subNodes);
        return accNodes;
    }, []);
    return nodes;
}
export function getEdges(node, visited = {}) {
    visited[node.key] = true;
    const subNodeEdges = Object.keys(node.states).reduce((edges, key) => {
        if (visited[key]) {
            return edges;
        }
        const subState = node.states[key];
        edges.push(...getEdges(subState, visited));
        visited[key] = true;
        return edges;
    }, []);
    if (!node.on) {
        return subNodeEdges;
    }
    const edges = Object.keys(node.on).reduce((accEdges, action) => {
        const subStateKey = node.on[action];
        const subNode = node.parent.getState(subStateKey);
        const edge = { action, source: node, target: subNode };
        accEdges.push(edge);
        if (!visited[subStateKey]) {
            accEdges.push(...getEdges(subNode, visited));
            visited[subStateKey] = true;
        }
        return accEdges;
    }, []);
    return subNodeEdges.concat(edges);
}
export function getAdjacencyMap(node) {
    const actionMap = node.parent ? {} : undefined;
    const adjacency = {};
    if (node.on) {
        for (const action of Object.keys(node.on)) {
            const nextState = node.parent.getState(node.on[action]);
            let nextStateId = nextState.id;
            if (nextState.initial) {
                nextStateId += '.' + nextState.initial;
            }
            actionMap[action] = nextStateId;
        }
    }
    if (actionMap) {
        adjacency[node.id] = actionMap;
    }
    if (node.states) {
        for (const stateKey of Object.keys(node.states)) {
            const state = node.states[stateKey];
            const stateAdjacency = mapValues(getAdjacencyMap(state), value => {
                return Object.assign({}, actionMap, value);
            });
            Object.assign(adjacency, stateAdjacency);
        }
    }
    return adjacency;
}
export function getShortestPaths(machine) {
    const adjacency = getAdjacencyMap(machine);
    const initialId = machine.states[machine.initial].id;
    const pathMap = {
        [initialId]: []
    };
    shortestPaths(adjacency, initialId, pathMap);
    return pathMap;
}
export function shortestPaths(adjacency, stateId, pathMap, visited = new Set()) {
    visited.add(stateId);
    const actionMap = adjacency[stateId];
    for (const action of Object.keys(actionMap)) {
        const nextStateId = actionMap[action];
        if (!pathMap[nextStateId] ||
            pathMap[nextStateId].length > pathMap[stateId].length + 1) {
            pathMap[nextStateId] = [...pathMap[stateId], stateId];
        }
    }
    for (const action of Object.keys(actionMap)) {
        const nextStateId = actionMap[action];
        if (visited.has(nextStateId)) {
            continue;
        }
        shortestPaths(adjacency, nextStateId, pathMap, visited);
    }
    return pathMap;
}
